import { describe, it, expect, beforeAll, test } from "@jest/globals";
import { JSDOM } from "jsdom";
import SigningService from "../../src/services/SigningService";
import { EMAIL_VENDORS } from "../../src/services/StatusesService";
import {
  arrayBufferToBase64,
  CryptoService,
  AttachmentsSignatures,
  AttachmentData,
} from "../../src";
import {
  findSeverityStateIndex,
  SEVERITY_LEVELS,
} from "../../src/services/SigningService/severityLevels";
import * as path from "path";
import * as fs from "fs";
import { stateExpectations } from "./severityLevels";

describe("SigningService tests", () => {
  let signingService: SigningService;
  let publicKeyPEM;
  let privateKeyPEM;

  beforeAll(async () => {
    const keys = await CryptoService.generateRSAKeys();
    publicKeyPEM = keys.publicKeyPEM;
    privateKeyPEM = keys.privateKeyPEM;

    signingService = new SigningService();

    /**
     * Must be defined by integrators.
     * @param value
     */
    signingService.helpers.signRSAExternal = async (value: string) => {
      return arrayBufferToBase64(
        await CryptoService.signRSA(
          privateKeyPEM,
          new TextEncoder().encode(value)
        )
      );
    };

    /**
     * Can be amended by integrator.
     * Test is using JSDOM library.
     */
    signingService.helpers.parseHtml = (htmlString) => {
      const jsdom = new JSDOM(htmlString);
      return jsdom.window.document;
    };
  });

  /**
   * https://code.vereign.com/light/documentation/-/blob/master/Validation.md#components-mode
   */
  describe("Components verification mode", () => {
    it("Checks a valid test case", async () => {
      const sentMime = fs
        .readFileSync(path.resolve(__dirname, "sent.eml"))
        .toString();
      const receivedMime = fs
        .readFileSync(path.resolve(__dirname, "received.eml"))
        .toString();

      /**
       * Signing logic
       */
      const {
        partsSignatures,
        rabinFingerprints,
      } = await signingService.signMimeComponents(
        EMAIL_VENDORS.OUTLOOK,
        sentMime
      );

      /**
       * Validation logic
       */
      const validationResults = await signingService.verifyMimeComponents(
        publicKeyPEM,
        EMAIL_VENDORS.OUTLOOK,
        receivedMime,
        partsSignatures,
        rabinFingerprints
      );

      expect(validationResults.htmlPartsSignatureValid).toBeTruthy();
      expect(validationResults.plainPartsSignatureValid).toBeTruthy();
      expect(validationResults.pseudoPlainPartsSignatureValid).toBeTruthy();
      expect(validationResults.attachmentsPartsSignatureValid).toBeTruthy();
      expect(validationResults.severityLevel).toEqual(SEVERITY_LEVELS.VALID);
    });
  });

  describe("Attachments signing/verification", () => {
    it("Checks a valid test case", async () => {
      const fileContentBytes = "abacadaba";
      const fileName = "abacadaba.txt";

      const hash = await signingService.helpers.cryptoService.MD5(
        fileContentBytes
      );

      const attachmentData: AttachmentData = {
        filename: fileName,
        type: "raw",
        content: fileContentBytes,
      };

      const attachmentHash: AttachmentData = {
        filename: fileName,
        type: "hash",
        content: arrayBufferToBase64(hash),
        hashingAlgorithm: "md5",
      };

      const attachments = [];
      attachments.push(attachmentData);
      attachments.push(attachmentHash);

      const signedAttachmentsData = await signingService.signAttachments(
        attachments
      );

      const signatures: AttachmentsSignatures = Object.keys(
        signedAttachmentsData
      ).reduce((acc, hash) => {
        const attachmentsData = signedAttachmentsData[hash];
        return Object.assign(acc, {
          [hash]: attachmentsData.map((attachmentData) => ({
            signature: attachmentData.signature,
            hashingAlgorithm: attachmentData.hashingAlgorithm,
          })),
        });
      }, {});

      let attachmentsToVerify = [];

      attachmentsToVerify.push(attachmentData);

      let verificationResult = await signingService.verifyAttachments(
        publicKeyPEM,
        attachmentsToVerify,
        signatures
      );

      expect(verificationResult.every((verified) => !!verified)).toBeTruthy();

      const attachmentData1 = {
        filename: fileName,
        type: "raw",
        content: "test",
      };

      attachmentsToVerify = [];
      attachmentsToVerify.push(attachmentData1);

      verificationResult = await signingService.verifyAttachments(
        publicKeyPEM,
        attachmentsToVerify,
        signatures
      );

      expect(verificationResult.every((verified) => !!verified)).toBeFalsy();

      const attachmentData2 = {
        filename: "test",
        type: "raw",
        content: "test",
      };

      attachmentsToVerify = [];

      attachmentsToVerify.push(attachmentData);
      attachmentsToVerify.push(attachmentData1);
      attachmentsToVerify.push(attachmentData2);
      attachmentsToVerify.push(attachmentHash);

      verificationResult = await signingService.verifyAttachments(
        publicKeyPEM,
        attachmentsToVerify,
        signatures
      );

      expect(verificationResult[0]).toBeTruthy();
      expect(verificationResult[1]).toBeFalsy();
      expect(verificationResult[2]).toBeFalsy();
      expect(verificationResult[3]).toBeTruthy();
    });
  });

  describe("Severity levels", () => {
    test.each(stateExpectations)(
      "Checks severity state %p",
      (state, stateIndex) => {
        const index = findSeverityStateIndex(
          state[0],
          state[1],
          state[2],
          state[3]
        );

        expect(stateIndex).toEqual(index);
      }
    );
  });
});
