import { HTMLNormalizer, PlainNormalizer } from "@vereign/mime-normalizer";
import PNG, { getMetaData } from "@vereign/png";
import MIMEParser from "@vereign/mime-parser";
import { ICryptoService } from "../CryptoService/ICryptoService";
import RKAUtility from "../../utils/rka";
import {
  arrayBufferToBase64,
  base64ToArrayBuffer,
  CryptoService,
} from "../../index";
import {
  AttachmentData,
  AttachmentsSignatures,
  PartsRabinFingerprints,
  PartsSignatures,
  SignComponentsResult,
  SignedAttachmentsData,
  SigningServiceHelpers,
  VerifyComponentsResult,
} from "./types";
import {
  findSeverityStateIndex,
  SEVERITY_LEVEL_PER_STATE,
  SEVERITY_MESSAGE_PER_STATE,
} from "./severityLevels";

class SigningService {
  public cryptoService: ICryptoService;
  public helpers: SigningServiceHelpers;

  constructor() {
    this.cryptoService = CryptoService;

    this.helpers = {
      signRSAExternal: async () => {
        throw new Error(
          "Please provide external function to calculate RSA signature"
        );
      },
      parseHtml: (htmlString) => {
        const parser = new DOMParser();
        return parser.parseFromString(htmlString, "text/html");
      },
      cryptoService: CryptoService,
      arrayBufferToBase64,
    };
  }

  /**
   * Function implementing COMPONENTS signing routine
   * https://code.vereign.com/light/documentation/-/blob/master/Validation.md#components-mode
   * @param vendor - email vendor (GMAIL | OUTLOOK | ROUNDCUBE | GENERIC_MIME)
   * @param mime - mime string
   */
  async signMimeComponents(
    vendor: string,
    mime: string | MIMEParser
  ): Promise<SignComponentsResult> {
    const { html, plain, attachmentsData } = await this.parseMime(mime);

    const document = this.helpers.parseHtml(html);
    const normalizedHTML = HTMLNormalizer.normalizeVendorHtml(document, vendor);
    const normalizedPlain = PlainNormalizer.normalizePlain(plain);
    const pseudoPlainPart = HTMLNormalizer.extractPseudoPlainPart(document);

    const hashedAttachments = await this.hashAttachments(
      attachmentsData,
      false
    );

    const [
      htmlPartsSignature,
      plainPartsSignature,
      pseudoPlainPartsSignature,
      attachmentsPartsSignature,
    ] = await this.getPartsSignatures([
      [normalizedHTML],
      [normalizedPlain],
      [pseudoPlainPart],
      hashedAttachments.map((attachment) => attachment.content),
    ]);

    const rabinFingerprints: PartsRabinFingerprints = {
      htmlPartRabinFingerprint: normalizedHTML
        ? RKAUtility.getRabinFingerprint(normalizedHTML)
        : "",
      plainPartRabinFingerprint: normalizedPlain
        ? RKAUtility.getRabinFingerprint(normalizedPlain)
        : "",
      pseudoPlainPartRabinFingerprint: pseudoPlainPart
        ? RKAUtility.getRabinFingerprint(pseudoPlainPart)
        : "",
      htmlPartSize: +(normalizedHTML && normalizedHTML.length),
      plainPartSize: +(normalizedHTML && normalizedPlain.length),
      pseudoPlainPartSize: +(pseudoPlainPart && pseudoPlainPart.length),
    };

    return {
      partsSignatures: {
        htmlPartsSignature,
        plainPartsSignature,
        pseudoPlainPartsSignature,
        attachmentsPartsSignature,
      },
      rabinFingerprints,
    };
  }

  async parseMime(
    mime: string | MIMEParser,
    ignoreQrCodes = true
  ): Promise<{
    html: string;
    plain: string;
    attachmentsData: AttachmentData[];
  }> {
    return new Promise((resolve) => {
      (async () => {
        let parsedMime;
        if (typeof mime === "string") {
          parsedMime = new MIMEParser(mime);
        } else if (mime instanceof MIMEParser) {
          parsedMime = mime;
        } else {
          throw new Error("Invalid MIME");
        }

        const html = await parsedMime.getHTML({
          populateAttachments: false,
          hashAttachment: async (base64) =>
            arrayBufferToBase64(await CryptoService.SHA256(base64, "base64")),
        });

        const plain = await parsedMime.getPlain({
          populateAttachments: false,
          hashAttachment: async (base64) =>
            arrayBufferToBase64(await CryptoService.SHA256(base64, "base64")),
        });

        const attachmentsParts = parsedMime.getAttachments();
        let attachmentsData: AttachmentData[] = attachmentsParts.map(
          (attachmentPart): AttachmentData => {
            const { base64, contentType } = parsedMime.getAttachment(
              attachmentPart
            );

            const filename = parsedMime.getFilenameFromHeaders(
              attachmentPart.headers
            );

            return {
              filename,
              content: base64,
              type: "raw",
              contentType,
            };
          }
        );

        if (ignoreQrCodes) {
          attachmentsData = attachmentsData.filter((attachmentData) => {
            if (attachmentData.contentType === "image/png") {
              const png = PNG.readPng(
                new Uint8Array(base64ToArrayBuffer(attachmentData.content))
              );
              const metaData = getMetaData(png);
              // Ignore QAR code PNGss
              // TODO: add flag "vereignQARCode" to the PNG meta data.
              return !(metaData.shash && metaData.timestamp);
            }
          });
        }

        resolve({
          html,
          plain,
          attachmentsData,
        });
      })();
    });
  }

  async getPartsSignatures(partsBatch: string[][]): Promise<string[]> {
    return await Promise.all(
      partsBatch.map(async (parts) => {
        const partsHash = await this.preparePartsForSigning(parts);
        if (partsHash) {
          return await this.helpers.signRSAExternal(partsHash);
        } else {
          return "";
        }
      })
    );
  }

  /**
   * Function implementing COMPONENTS signing routine
   * https://code.vereign.com/light/documentation/-/blob/master/Validation.md#components-mode
   * @param publicKeyPEM
   * @param vendor - email vendor (GMAIL | OUTLOOK | ROUNDCUBE | GENERIC_MIME)
   * @param mime - MIME string
   * @param partsSignatures
   * @param partsRabinFingerprints
   */
  async verifyMimeComponents(
    publicKeyPEM: string,
    vendor: string,
    mime: string | MIMEParser,
    partsSignatures: PartsSignatures,
    partsRabinFingerprints: PartsRabinFingerprints
  ): Promise<VerifyComponentsResult> {
    const { html, plain, attachmentsData } = await this.parseMime(mime);

    const document = this.helpers.parseHtml(html);
    const normalizedHTML = HTMLNormalizer.normalizeVendorHtml(document, vendor);
    const normalizedPlain = PlainNormalizer.normalizePlain(plain);
    const pseudoPlainPart = HTMLNormalizer.extractPseudoPlainPart(document);
    const hashedAttachments = await this.hashAttachments(
      attachmentsData,
      false
    );

    const originalNormalizedHtml = RKAUtility.findRKASubstring(
      normalizedHTML,
      partsRabinFingerprints.htmlPartRabinFingerprint,
      partsRabinFingerprints.htmlPartSize
    );

    const originalNormalizedPlain = RKAUtility.findRKASubstring(
      normalizedPlain,
      partsRabinFingerprints.plainPartRabinFingerprint,
      partsRabinFingerprints.plainPartSize
    );

    const originalPseudoPlain = RKAUtility.findRKASubstring(
      pseudoPlainPart,
      partsRabinFingerprints.pseudoPlainPartRabinFingerprint,
      partsRabinFingerprints.pseudoPlainPartSize
    );

    const [
      htmlPartsValid,
      plainPartsValid,
      pseudoPlainPartsValid,
      attachmentsPartsValid,
    ] = await this.verifyPartsSignatures(publicKeyPEM, [
      {
        signature: partsSignatures.htmlPartsSignature,
        parts: [originalNormalizedHtml],
      },
      {
        signature: partsSignatures.plainPartsSignature,
        parts: [originalNormalizedPlain],
      },
      {
        signature: partsSignatures.pseudoPlainPartsSignature,
        parts: [originalPseudoPlain],
      },
      {
        signature: partsSignatures.attachmentsPartsSignature,
        parts: hashedAttachments.map((attachment) => attachment.content),
      },
    ]);

    const severityStateIndex = findSeverityStateIndex(
      htmlPartsValid,
      plainPartsValid,
      pseudoPlainPartsValid,
      attachmentsPartsValid
    );

    return {
      htmlPartsSignatureValid: htmlPartsValid,
      plainPartsSignatureValid: plainPartsValid,
      pseudoPlainPartsSignatureValid: pseudoPlainPartsValid,
      attachmentsPartsSignatureValid: attachmentsPartsValid,
      severityLevel: SEVERITY_LEVEL_PER_STATE[severityStateIndex],
      statusMessage: SEVERITY_MESSAGE_PER_STATE[severityStateIndex],
    };
  }

  async verifyPartsSignatures(
    publicKeyPEM: string,
    partsBatch: { signature: string; parts: string[] }[]
  ): Promise<boolean[]> {
    return await Promise.all(
      partsBatch.map(async ({ signature, parts }) => {
        const partsHash = await this.preparePartsForSigning(parts);

        if (partsHash) {
          return await this.helpers.cryptoService.verifyRSASignature(
            publicKeyPEM,
            Buffer.from(partsHash),
            Buffer.from(signature, "base64")
          );
        } else {
          // If there's no hash, but signature exists, then it's invalid
          return !signature;
        }
      })
    );
  }

  /**
   * Function implementing specification
   * https://code.vereign.com/light/documentation/-/blob/master/Validation.md#partssignaturealgorithm
   *
   * @param parts - array of strings representing parts
   */
  async preparePartsForSigning(parts: Array<string>): Promise<string> {
    const hashedParts = await Promise.all(
      parts.map(async (part) => {
        return await this.helpers.cryptoService.SHA256(part);
      })
    );

    return hashedParts
      .map((hash) => this.helpers.arrayBufferToBase64(hash))
      .sort()
      .join("\n");
  }

  async hashAttachments(
    attachments: AttachmentData[],
    useHashingAlgorithm: boolean
  ): Promise<AttachmentData[]> {
    const result = [];

    const cryptoService = this.helpers.cryptoService;
    const hashingFunctions = {
      sha1: cryptoService.SHA1,
      sha256: cryptoService.SHA256,
      sha384: cryptoService.SHA384,
      sha512: cryptoService.SHA512,
      md5: cryptoService.MD5,
    };

    for (let i = 0; i < attachments.length; i++) {
      const attachment = attachments[i];

      if (attachment.type == "raw") {
        let hash;
        let hashingAlgorithm;

        if (useHashingAlgorithm == true) {
          hashingAlgorithm = attachment.hashingAlgorithm;
          const hashingFunction = hashingFunctions[hashingAlgorithm];

          if (hashingFunction) {
            hash = await hashingFunction(attachment.content);
          } else {
            throw new Error(
              `No such hashing algorithm supported: "${hashingAlgorithm}"`
            );
          }
        } else {
          hashingAlgorithm = "sha256";
          hash = await this.helpers.cryptoService.SHA256(attachment.content);
        }

        const attachmentHashed = {
          ...attachment,
          content: hash,
          type: "hash",
          hashingAlgorithm: hashingAlgorithm,
        };

        result.push(attachmentHashed);
      } else {
        result.push(attachment);
      }
    }

    return result;
  }

  /**
   * @param attachments - an array of attachments data objects
   */
  async signAttachments(
    attachments: AttachmentData[]
  ): Promise<SignedAttachmentsData> {
    const hashedAttachments = await this.hashAttachments(attachments, false);

    const result = {};

    for (let i = 0; i < hashedAttachments.length; i++) {
      const hashedAttachment = hashedAttachments[i];

      const wrappingHash = await this.helpers.cryptoService.SHA256(
        hashedAttachment.content
      );
      const filenameHashBuffer = await this.helpers.cryptoService.SHA256(
        hashedAttachment.filename
      );
      const filenameHash = this.helpers.arrayBufferToBase64(filenameHashBuffer);

      const signature = await this.helpers.signRSAExternal(
        this.helpers.arrayBufferToBase64(wrappingHash)
      );

      if (!(filenameHash in result)) {
        result[filenameHash] = [];
      }

      hashedAttachment.signature = signature;

      result[filenameHash].push(hashedAttachment);
    }

    return result;
  }

  /**
   * @param publicKeyPEM
   * @param attachments - an array of attachments hashes
   * @param signatures
   */
  async verifyAttachments(
    publicKeyPEM: string,
    attachments: AttachmentData[],
    signatures: AttachmentsSignatures
  ): Promise<boolean[]> {
    const result = [];

    for (let i = 0; i < attachments.length; i++) {
      const attachment = attachments[i];
      const filename = attachment.filename;
      const filenameHashBuffer = await this.helpers.cryptoService.SHA256(
        filename
      );
      const filenameHash = this.helpers.arrayBufferToBase64(filenameHashBuffer);

      if (filenameHash in signatures) {
        const attachmentSignatures = signatures[filenameHash];

        let attachmentVersionsHashed = [];

        if (attachment.type == "raw") {
          const hashingAlgorithms = {};

          for (let j = 0; j < attachmentSignatures.length; j++) {
            const signature = attachmentSignatures[j];
            hashingAlgorithms[signature.hashingAlgorithm] = true;
          }

          const attachmentVersions = [];

          for (const hashingAlgorithm in hashingAlgorithms) {
            if (hashingAlgorithms[hashingAlgorithm]) {
              const attachmentVersion = {
                filename: filename,
                content: attachment.content,
                type: "raw",
                hashingAlgorithm: hashingAlgorithm,
              };

              attachmentVersions.push(attachmentVersion);
            }
          }

          attachmentVersionsHashed = await this.hashAttachments(
            attachmentVersions,
            true
          );
        } else {
          attachmentVersionsHashed.push(attachment);
        }

        let valid = false;

        for (let j = 0; j < attachmentSignatures.length; j++) {
          if (valid === true) {
            break;
          }
          const signature = attachmentSignatures[j];
          for (let k = 0; k < attachmentVersionsHashed.length; k++) {
            const attachmentHash = attachmentVersionsHashed[k];
            const wrappingHash = await this.helpers.cryptoService.SHA256(
              attachmentHash.content
            );
            const base64WrappingHash = this.helpers.arrayBufferToBase64(
              wrappingHash
            );

            const isValid = await this.helpers.cryptoService.verifyRSASignature(
              publicKeyPEM,
              Buffer.from(base64WrappingHash),
              Buffer.from(signature.signature, "base64")
            );

            if (isValid == true) {
              valid = true;
              break;
            }
          }
        }

        result.push(valid);
      } else {
        result.push(false);
      }
    }

    return result;
  }
}

export default SigningService;
