import { CryptoService } from "../../index";

const EventEmitter = require("eventemitter2");
import { MerkleTree as MerkleTreeJS } from "merkletreejs";
import AeternityService from "../AeternityService/AeternityService";
import CloudflareService from "../CloudflareService";
import {
  BlockData,
  MerkleTree,
  MerkleTreeVerificationDetails,
  StatusData,
  TxData,
  VerificationData,
} from "../../types";
import VerificationError from "./VerificationError";
import { AxiosError } from "axios";

export const TRANSACTION_RETRIEVED = "TRANSACTION_RETRIEVED";
export const MERKLE_TREE_VERIFIED = "MERKLE_TREE_VERIFIED";
export const BLOCK_DATA_RETRIEVED = "BLOCK_DATA_RETRIEVED";

class VerificationService extends EventEmitter {
  private _aeternityService: AeternityService = null;
  private _cloudflareService: CloudflareService = null;

  constructor(
    cdnUrl?: string,
    cdnBucket?: string,
    aeternityNodeUrls?: string[] | string,
    aeternityCompilerUrls?: string[] | string,
    aeternityContractBytecode?: string
  ) {
    super();

    this._aeternityService = new AeternityService(
      aeternityNodeUrls,
      aeternityCompilerUrls,
      aeternityContractBytecode
    );

    this._cloudflareService = new CloudflareService(cdnUrl, cdnBucket);
  }

  async getTransaction(hash: string): Promise<TxData> {
    return this._aeternityService.getTxDataByHash(hash).catch((err) => {
      this.catchVerificationError(err, `Error getting transaction`);
      return null;
    });
  }

  async getBlockData(blockHeight: number): Promise<BlockData> {
    return this._aeternityService.getBlockData(blockHeight).catch((e) => {
      this.catchVerificationError(e, `Error retrieving block`);
      return null;
    });
  }

  async verifyStatusData(statusData: StatusData): Promise<VerificationData> {
    let verificationError: string;
    let blockData: BlockData;
    let merkleTreeDetails: MerkleTreeVerificationDetails;
    let txData: TxData;

    try {
      txData = await this.getTransaction(statusData.transactionHash);
      this.emit(TRANSACTION_RETRIEVED, txData);

      const merkleTree = await this.getMerkleTreeForTransaction(txData);

      const statusVerified = await this.verifyMerkleTreeNode(
        statusData.statusRaw,
        merkleTree
      );

      merkleTreeDetails = {
        verified: statusVerified,
        treeSize: merkleTree.nodes.length,
        statusPosition: merkleTree.nodes.indexOf(statusData.statusRaw) + 1,
        rootHash: merkleTree.rootHash,
      };

      this.emit(MERKLE_TREE_VERIFIED, merkleTreeDetails);

      blockData = await this.getBlockData(txData.blockHeight);

      this.emit(BLOCK_DATA_RETRIEVED, blockData);
    } catch (e) {
      if (e instanceof VerificationError) {
        verificationError = e.message;
      } else {
        throw e;
      }
    }

    return {
      statusData,
      merkleTreeDetails,
      blockData,
      verificationError,
      txData,
    };
  }

  async getMerkleTreeForTransaction(txData: TxData): Promise<MerkleTree> {
    try {
      const txMerkleTreeData = await this._aeternityService.decodeContractCallData(
        txData.tx?.callData
      );

      const merkleTreeNodes = await this._cloudflareService.getMerkleTree(
        txMerkleTreeData.merkleeTreeFileName
      );

      return {
        nodes: merkleTreeNodes,
        rootHash: txMerkleTreeData.rootNodeHash,
      };
    } catch (e) {
      this.catchVerificationError(
        e,
        `Error extracting Merkle tree from transaction.`
      );
    }
  }

  async verifyMerkleTreeNode(
    nodeToVerify: string,
    merkleTree: MerkleTree
  ): Promise<boolean> {
    const leaves = await Promise.all(
      merkleTree.nodes.map(async (x) =>
        Buffer.from(await CryptoService.SHA256(x))
      )
    );

    const tree = new MerkleTreeJS(leaves);

    const leaf = await CryptoService.SHA256(nodeToVerify);
    const proof = tree.getProof(
      Buffer.from(leaf),
      merkleTree.nodes.indexOf(nodeToVerify)
    );

    const verified = tree.verify(
      proof,
      Buffer.from(leaf),
      Buffer.from(merkleTree.rootHash, "base64")
    );

    if (!verified) {
      throw new VerificationError(`Merkle tree not verified.`);
    }

    return verified;
  }

  catchVerificationError(
    error: AxiosError,
    verificationFailureMessage: string
  ): void {
    if (error.response) {
      throw new VerificationError(verificationFailureMessage);
    } else {
      throw error;
    }
  }
}

VerificationService.TRANSACTION_RETRIEVED = TRANSACTION_RETRIEVED;
VerificationService.MERKLE_TREE_VERIFIED = MERKLE_TREE_VERIFIED;
VerificationService.BLOCK_DATA_RETRIEVED = BLOCK_DATA_RETRIEVED;

export default VerificationService;
